//---------------------------------------------------------------------------
// Created 27-Sep-2003: Thomas Mulgrew (tmulgrew@slingshot.co.nz)
// Copyright (C) Thomas Mulgrew
#include <vcl.h>
#pragma hdrstop

#include <registry.hpp>
#include "Basic4GLv2Main.h"                 
#include "gbAbout.h"
#include "TomWindowsBasicLib.h"
#include "TomStdBasicLib.h"
#include "TomOpenGLBasicLib.h"
#include "GLBasicLib_gl.h"
#include "GLBasicLib_glu.h"
#include "TomTextBasicLib.h"
#include "glBasicConfigMain.h"
#include "VirtualMachineView.h"
#include "FunctionReferenceDlg.h"
#include "Misc.h"
#include "CreateStandalone.h"
#include "TomJoystickBasicLib.h"
#include "TomTrigBasicLib.h"
#include "TomFileIOBasicLib.h"
//#include "TomNetBasicLib.h"

//---------------------------------------------------------------------------
#pragma link "MemoEx"
#pragma resource "*.dfm"

const int gbStepsUntilRefresh = 1000;           // Process windows messages every X virtual machine steps

char TBasic4GLv2MainForm::ReadChar (TFileStream& s) {
    char c;
    do {
        if (s.Read (&c, sizeof (c)) == 0)
            c = 0;
    } while (c == 10);              // Skip newlines
    return c;
}

AnsiString TBasic4GLv2MainForm::ReadString (TFileStream& s) {
    AnsiString result = "";
    char c = ReadChar (s);
    while (c != 13 && c != 0) {
        result = result + c;
        c = ReadChar (s);
    }
    return result;
}

char *eol = "\r\n";
void TBasic4GLv2MainForm::WriteString (TFileStream& s, AnsiString str, bool addEol) {
    s.Write (str.c_str (), str.Length ());      // Write string
    if (addEol)
        s.Write (eol, 2);                       // Write newline
}

TBasic4GLv2MainForm *Basic4GLv2MainForm;
//---------------------------------------------------------------------------
__fastcall TBasic4GLv2MainForm::TBasic4GLv2MainForm(TComponent* Owner)
    : m_comp (m_vm), TForm(Owner)
{
    m_mode              = AP_STOPPED;
    m_vmView            = new TVirtualMachineViewForm (this, m_comp);
    m_filename          = "";
    m_modified          = false;
    m_debugMode         = false;
    m_delayScreenSwitch = false;
    m_startDirectory    = "";
    RefreshActions ();
    RefreshDebugDisplays ();
}
//---------------------------------------------------------------------------
void __fastcall TBasic4GLv2MainForm::AboutActionExecute(TObject *Sender)
{
    AboutBox->ShowModal ();
}
//---------------------------------------------------------------------------
void __fastcall TBasic4GLv2MainForm::FormCreate(TObject *Sender)
{
    Application->HintPause = 0;
    Application->HintHidePause = 0x7ffff;


    // Force load/save dialogs to open in current directory
    ForceCurrentDirectory = true;

    // Load settings from registry
    int width, height, bpp;
    bool fullScreen, border;
    ResetGLModeType resetMode;
    LoadRegistrySettings ();
    LoadScreenRegistrySettings (width, height, bpp, fullScreen, border, resetMode); 

    // Initialise application
    m_glWin     = NULL;
    m_glText    = NULL;

    // Create OpenGL window
    m_glWin = new glTextGridWindow (fullScreen, border, width, height, bpp, "Basic4GL", resetMode);
    if (m_glWin->Error ()) {
        MessageDlg (    (AnsiString) m_glWin->GetError ().c_str () + "\r\n\r\n"
                    +   "Please select a video mode that your graphics system supports.\r\n"
                    +   "You will then need to restart Basic4GL.",
                    mtError,
                    TMsgDlgButtons() << mbOK,
                    0);
        OptionsActionExecute (this);
        Application->Terminate ();
        return;
    }
    m_glWin->Hide ();

    // Create OpenGL text grid
    m_glText = new glSpriteEngine ((ExtractFilePath (Application->ExeName) + "charset.png").c_str (), &m_files, 25, 40, 16, 16);
    if (m_glText->Error ()) {
        MessageDlg (    (AnsiString) + m_glText->GetError ().c_str () + "\r\n\r\n"
                    +   "Basic4GL will close.",
                    mtError,
                    TMsgDlgButtons() << mbOK,
                    0);
        Application->Terminate ();
        return;
    }
    m_glWin->SetTextGrid (m_glText);

    // Plug in constant and function libraries
    InitTomStdBasicLib (m_comp);                // Standard library
    InitTomWindowsBasicLib (m_comp, &m_files);  // Windows specific library
    InitTomOpenGLBasicLib (m_comp, m_glWin, &m_files);    // OpenGL
    InitTomTextBasicLib (m_comp, m_glWin, m_glText);
    InitGLBasicLib_gl (m_comp);
    InitGLBasicLib_glu (m_comp);
    InitTomJoystickBasicLib (m_comp, m_glWin);
    InitTomTrigBasicLib (m_comp);               // Trigonometry library
    InitTomFileIOBasicLib (m_comp, &m_files);   // File I/O library
//    InitTomNetBasicLib (m_comp);

    // Clear text
    NewActionExecute (this);

    // Default directory = programs. (This allows copy/pasted programs to find texture and data files without having to be saved first.) 
    SetCurrentDir (m_startDirectory);

    // Setup text highlighting styles
    ResetStyles ();
}
//---------------------------------------------------------------------------
void __fastcall TBasic4GLv2MainForm::SaveAsActionExecute(TObject *Sender)
{
    SaveAs ();
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::NewActionExecute(TObject *Sender)
{

    // Check if modified
    if (!CheckModified ())
        return;

    // Clear source code
    SourceCode->Clear();
    SourceCode->SelStart = 0;
    m_filename  = "";
    m_modified   = false;
    m_vm.ClearUserBreakPts ();
    RefreshActions ();
}
//---------------------------------------------------------------------------
bool TBasic4GLv2MainForm::CheckModified () {

    // Can proceed if not modified
    if (!m_modified)
        return true;

    // Otherwise prompt user whether to save
    switch (MessageDlg (    "Save Changes?",
                            mtConfirmation,
                            TMsgDlgButtons() << mbYes << mbNo << mbCancel,
                            0)) {
    case mrYes:     return Save ();     // Proceed if save is successful
    case mrNo:      return true;        // Don't save, but proceed anyway
    case mrCancel:  return false;       // Don't save or proceed
    }
    return false;
}
//---------------------------------------------------------------------------
bool TBasic4GLv2MainForm::Save () {

    // If filename not specified yet, do save as instead
    if (m_filename == "")
        return SaveAs ();

    // Create backup file first
    try {
        TFileStream *source, *backup;
        try         { source = new TFileStream (m_filename, fmOpenRead); }
        catch (...) { source = NULL; }

        if (source != NULL) {
            try {
                AnsiString filename = ChangeFileExt (m_filename, ".~gb");
                try         {   backup = new TFileStream (filename, fmOpenWrite); }
                catch (...) {   backup = new TFileStream (filename, fmCreate); }

                try {
                    backup->Size = 0;
                    backup->CopyFrom (source, source->Size);
                }
                __finally {
                    delete backup;
                }
            }
            __finally {
                delete source;
            }
        }
    }
    catch (...) {;}

    // Then save the file
    try {

        // Save file to disk
        TFileStream *file;
        try         {   file = new TFileStream (m_filename, fmOpenWrite); }
        catch (...) {   file = new TFileStream (m_filename, fmCreate); }
        try {

            // Clear file
            file->Size = 0;
            
            // Write out each line
            for (int i = 0; i < SourceCode->Lines->Count; i++)
                WriteString (*file, SourceCode->Lines->Strings [i], i < SourceCode->Lines->Count - 1);

            // Update state
            m_modified = false;
            (*StatusBar->Panels) [2]->Text = m_filename + " saved";
        }
        __finally {
            delete file;
            RefreshActions ();
        }
    }
    catch (const Exception &e) {
        MessageDlg (e.Message, mtInformation, TMsgDlgButtons () << mbOK, 0);
        return false;
    }
    return true;
}
//---------------------------------------------------------------------------
bool TBasic4GLv2MainForm::SaveAs () {
    SaveDialog->FileName = m_filename;
    if (SaveDialog->Execute () && SaveDialog->FileName != "") {
        m_filename = SaveDialog->FileName;
        return Save ();
    }
    else
        return false;
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::SaveActionExecute(TObject *Sender)
{
    Save ();    
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::FormClose(TObject *Sender,
      TCloseAction &Action)
{
    // Stop program running
    if (m_mode == AP_RUNNING || m_mode == AP_PAUSED) {
        m_mode = AP_STOPPED;
        RefreshActions ();
        Action = caNone;
        return;
    }

    // Save file before closing
    if (!CheckModified ())
        Action = caNone;        // User cancelled

    // Library cleanup functions
    ShutDownTomWindowsBasicLib ();
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::LoadActionExecute(TObject *Sender)
{

    // Save current file
    if (!CheckModified ())
        return;

    // Load new one
    if (OpenDialog->Execute ()) {

        // Load text
        TFileStream *file = new TFileStream (OpenDialog->FileName, fmOpenRead);
        SourceCode->Hide ();
        try {
            SourceCode->CaretX = 0;
            SourceCode->CaretY = 0;
            SourceCode->Clear ();
            while (file->Position < file->Size)
                SourceCode->Lines->Add (ReadString (*file));

            // Update state
            m_filename = OpenDialog->FileName;
            m_modified = false;
        }
        __finally {
            delete file;
            SourceCode->Show ();
            PlaceCursor (0, 0);
            m_vm.ClearUserBreakPts ();
            RefreshActions ();
        }
    }
}
//---------------------------------------------------------------------------
void __fastcall TBasic4GLv2MainForm::SourceCodeChange(TObject *Sender)
{
    m_modified = true;
    RefreshActions ();
}
//---------------------------------------------------------------------------
void TBasic4GLv2MainForm::RefreshActions () {

    // Enable/disable actions to reflect state
    switch (m_mode) {
    case AP_STOPPED:
        OptionsAction->Enabled  = true;
        SourceCode->ReadOnly    = false;
        SaveAction->Enabled     = m_modified;
        RunAction->ImageIndex   = 0;
        LoadAction->Enabled     = true;
        SaveAsAction->Enabled   = true;
        NewAction->Enabled      = true;
        CreateStandAlone->Enabled   = true;
        break;
    case AP_RUNNING:
        OptionsAction->Enabled  = false;
        SourceCode->ReadOnly    = true;
        SaveAction->Enabled     = false;
        RunAction->ImageIndex   = 4;
        LoadAction->Enabled     = false;
        SaveAsAction->Enabled   = false;
        NewAction->Enabled      = false;
        CreateStandAlone->Enabled   = false;
        break;
    case AP_PAUSED:
        OptionsAction->Enabled  = false;
        SourceCode->ReadOnly    = true;
        SaveAction->Enabled     = false;
        RunAction->ImageIndex   = 4;
        LoadAction->Enabled     = false;
        SaveAsAction->Enabled   = false;
        NewAction->Enabled      = false;
        CreateStandAlone->Enabled   = false;
        break;
    };

    if (m_filename == "") {
        Caption = "Basic4GL - Copyright (C) Tom Mulgrew 2004";
        Application->Title = "Basic4GL";
    }
    else {
        Caption = ExtractFileName (m_filename) + " - Basic4GL";
        Application->Title = Caption;
    }
}

//---------------------------------------------------------------------------
void TBasic4GLv2MainForm::LoadProgramIntoCompiler () {
    m_comp.Parser ().SourceCode ().clear ();
    for (int i = 0; i < SourceCode->Lines->Count; i++)
        m_comp.Parser ().SourceCode().push_back(SourceCode->Lines->Strings [i].c_str ());
}

//---------------------------------------------------------------------------
bool TBasic4GLv2MainForm::Compile () {

    // Compile the program, and reset the IP to the start.
    // Returns true if program compiled successfully. False if compiler error occurred.
    m_mode = AP_STOPPED;

    // Compile
    LoadProgramIntoCompiler ();
    m_comp.ClearError ();
    m_comp.Compile ();
    if (m_comp.Error ()) {
        PlaceCursor (m_comp.Line(), m_comp.Col ());
        (*StatusBar->Panels) [2]->Text = m_comp.GetError ().c_str ();
        return false;
    }

    // Reset Virtual machine
    m_vm.Reset ();

    // Reset OpenGL state
//    m_glWin->InitGL ();
    m_glWin->ResetGL ();
    m_glWin->Activate ();
    m_glWin->OpenGLDefaults ();

    return true;
}

//---------------------------------------------------------------------------
void TBasic4GLv2MainForm::Continue () {

    // Resume running the current program

    // Set running state
    (*StatusBar->Panels) [2]->Text = "Running...";
    m_mode = AP_RUNNING;
    RefreshActions ();
    RefreshDebugDisplays ();

    // Show and activate OpenGL window
    int counter = 0;
    if (!m_delayScreenSwitch)
        m_glWin->Activate ();
    else
        counter = 2;            // Activate screen second time around main loop.

    // Run loop
    try {
        // Kick the virtual machine over the next op-code before patching in the breakpoints.
        // otherwise we would never get past a breakpoint once we hit it, because we would
        // keep on hitting it immediately and returning.
        m_vm.Continue (1);
        do {

            if (counter > 0 && --counter == 0)
                m_glWin->Activate ();

            // Run the virtual machine for a certain number of steps
            m_vm.PatchIn ();
            m_vm.Continue (gbStepsUntilRefresh);

            // Process windows messages (to keep application responsive)
            Application->ProcessMessages ();
            m_glWin->ProcessWindowsMessages();
            if (m_glWin->PausePressed ())           // Check for pause key. (This allows us to pause when in full screen mode. Useful for debugging.)
                m_vm.Pause ();
        } while (       m_mode == AP_RUNNING
                    &&  !m_vm.Error ()
                    &&  !m_vm.Done ()
                    &&  !m_vm.Paused ()
                    &&  !m_glWin->Closing ());
    }
    catch (...) {
        // Basic4GL is designed not to throw exceptions.
        // But just to be sure...
        m_vm.MiscError ("An exception occured");
    }

    // Determine whether we are paused or stopped. (If we are paused, we can
    // resume from the current position. If we are stopped, we cannot.)
    if (m_vm.Paused () && !m_vm.Error () && !m_vm.Done () && !m_glWin->Closing ())
        m_mode = AP_PAUSED;
    else
        m_mode = AP_STOPPED;
    RefreshActions ();
    RefreshDebugDisplays ();

    // Force redraw of source code gutter
    SourceCode->InvalidateGutter ();
    SourceCode->Invalidate ();

    // Clear temp breakpoints (user ones still remain)
    // This also patches out all breakpoints.
    m_vm.ClearTempBreakPts ();

    // Handle GL window
    if (m_glWin->Closing ())                // Explicitly closed
        m_glWin->Hide ();                   // Hide it

    else {
        if (m_glWin->FullScreen () && m_glWin->Visible ()) {
            if (m_vm.Done ()) {             // Program ended normally

                // Wait for keypress.
                m_glWin->ProcessWindowsMessages();
                while (m_glWin->GetKey () != 0) m_glWin->ProcessWindowsMessages();
                while (m_glWin->GetKey () == 0) m_glWin->ProcessWindowsMessages();
            }

            m_glWin->Hide ();
        }
    }
    m_glWin->SetClosing (false);

    // Get focus back
    if (!(m_glWin->Visible () && !m_glWin->FullScreen () && m_vm.Done ())) {  // If program ended cleanly in windowed mode, leave focus on OpenGL window
        Show ();
        Activate ();
        if (SourceCode->CanFocus ())
            try { SourceCode->SetFocus (); }
            catch (...) { ; }
    }

    // Place cursor on current instruction
    if (m_vm.Error () || m_mode == AP_PAUSED) {
        int line, col;
        m_vm.GetIPInSourceCode (line, col);
        PlaceCursor (line, col);
    }

    // Update status bar
    if (m_vm.Error ())
        (*StatusBar->Panels) [2]->Text = m_vm.GetError ().c_str ();
    else if (m_mode == AP_STOPPED) {
        if (m_vm.Done ())   (*StatusBar->Panels) [2]->Text = "Program completed";
        else                (*StatusBar->Panels) [2]->Text = "Program stopped";
    }
    else
        (*StatusBar->Panels) [2]->Text = "Program paused. Click play button to resume.";
}

//---------------------------------------------------------------------------
void __fastcall TBasic4GLv2MainForm::RunActionExecute(TObject *Sender)
{
    if (m_mode == AP_STOPPED) {

        // Compile and run program from start
        if (Compile ())
            Continue ();
    }
    else  {

        // Stop program completely.
        m_mode = AP_STOPPED;
        RefreshActions ();
        RefreshDebugDisplays ();
    }
}
//---------------------------------------------------------------------------
void TBasic4GLv2MainForm::PlaceCursor (int line, int col) {

    // Put focus in sourcecode window
    if (SourceCode->CanFocus ())
        try { SourceCode->SetFocus (); }
        catch (...) { ; }

    SourceCode->CaretY = line;
    SourceCode->CaretX = col;
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::FormDestroy(TObject *Sender)
{
    // Free virtual machine resources
    // It's a good idea to do this first, while the OpenGL context and window
    // are still active.
    m_vm.Clr ();

    // Call library cleanup functions
//    CleanupTomNetBasicLib ();

    // Free text display
    if (m_glText != NULL)
        delete m_glText;

    // Free the openGL window
    if (m_glWin != NULL)
        delete m_glWin;

    // Save registry settings
    try { SaveRegistrySettings (); }
    catch (...) { ; }
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::ApplicationEvents1Idle(TObject *Sender,
      bool &Done)
{
    m_glWin->ProcessWindowsMessages ();

    // Hide window if user closes it (and program is not running)
    if (m_mode != AP_RUNNING && m_glWin->Closing ()) {
        m_glWin->Hide ();
        m_glWin->SetClosing (false);
    }

    static int lastX = 0, lastY = 0;
    int x = SourceCode->CaretX, y = SourceCode->CaretY;
    if (x != lastX || y != lastY) {
        lastX = x;
        lastY = y;
        (*StatusBar->Panels) [0]->Text = IntToStr (lastY + 1);
        (*StatusBar->Panels) [1]->Text = IntToStr (lastX + 1);
    }
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::OptionsActionExecute(TObject *Sender)
{

    // Show options dialog
    TConfigDialog *dlg = new TConfigDialog (this);
    try {
        if (dlg->ShowModal () == mrOk) {

            // Recreate the output window with the new settings
            int width, height, bpp;
            bool fullScreen, border;
            ResetGLModeType resetMode;
            LoadScreenRegistrySettings (width, height, bpp, fullScreen, border, resetMode);
            m_glWin->RecreateWindow(fullScreen, border, width, height, bpp, m_glWin->Title ());
            m_glWin->ResetGLMode() = resetMode;
        }
    }
    __finally {
        delete dlg;
    }
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::VMActionExecute(TObject *Sender)
{

    // Show virtual machine
    m_vmView->UpdateData ();
    m_vmView->ShowModal ();
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::ShowFunctionsActionExecute(
      TObject *Sender)
{
    FunctionRefDlg->Populate (m_comp);
    FunctionRefDlg->Show ();
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::ColorSelect1Accept(TObject *Sender)
{
    SourceCode->Color = ColorSelect1->Dialog->Color;
    ResetStyles ();
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::ColorSelect1BeforeExecute(
      TObject *Sender)
{
    ColorSelect1->Dialog->Color = SourceCode->Color;
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::FontEdit1BeforeExecute(TObject *Sender)
{
    FontEdit1->Dialog->Font = SourceCode->Font;
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::FontEdit1Accept(TObject *Sender)
{
    SourceCode->Font = FontEdit1->Dialog->Font;
    ResetStyles ();
}
//---------------------------------------------------------------------------

void TBasic4GLv2MainForm::LoadRegistrySettings () {

    // Load settings from registry
    try {
        TRegistry *Reg = new TRegistry ();
        try
        {
            Reg->RootKey = HKEY_CURRENT_USER;
            if (Reg->OpenKeyReadOnly("\\Software\\Tom\\Basic4GLv2")) {
                SourceCode->Color       = Reg->ReadInteger ("Colour");
                SourceCode->Font->Name  = Reg->ReadString  ("FontName");
                SourceCode->Font->Size  = Reg->ReadInteger ("FontSize");
                SourceCode->Font->Color = Reg->ReadInteger ("FontColor");
                m_startDirectory        = Reg->ReadString ("StartDirectory");
                if (m_startDirectory == "")
                    m_startDirectory = "Programs";
            }
        }
        __finally
        {
            delete Reg;
        }
    } catch (...) { ; }

}
//---------------------------------------------------------------------------
void TBasic4GLv2MainForm::SaveRegistrySettings () {

    // Save settings from registry
    TRegistry *Reg = new TRegistry;
    try
    {
        Reg->RootKey = HKEY_CURRENT_USER;
        if (Reg->OpenKey("\\Software\\Tom\\Basic4GLv2", true)) {
            Reg->WriteInteger ("Colour",    SourceCode->Color      );
            Reg->WriteString  ("FontName",  SourceCode->Font->Name );
            Reg->WriteInteger ("FontSize",  SourceCode->Font->Size );
            Reg->WriteInteger ("FontColor", SourceCode->Font->Color);
            Reg->WriteString  ("StartDirectory", GetCurrentDir ());
        }
    }
    __finally
    {
        delete Reg;
    }
}
//---------------------------------------------------------------------------

void ShowDoc (char *name) {
    ShellExecute (  Application->Handle,
                    "open",
                    name,
                    NULL,
                    ExtractFilePath (Application->ExeName).c_str (),
                    0);
}

void __fastcall TBasic4GLv2MainForm::GettingStarted1Click(TObject *Sender)
{
    ShowDoc ("Getting Started.htm");
}
//---------------------------------------------------------------------------


void __fastcall TBasic4GLv2MainForm::ProgrammersGuide1Click(TObject *Sender)
{
    ShowDoc ("Programmer's guide.htm");
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::LanguageGuide1Click(TObject *Sender)
{
    ShowDoc ("Language guide.htm");    
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::OpenGLGuide1Click(TObject *Sender)
{
    ShowDoc ("OpenGL guide.htm");
}
//---------------------------------------------------------------------------
void __fastcall TBasic4GLv2MainForm::SpriteLibraryGuide1Click(
      TObject *Sender)
{
    ShowDoc ("Sprite Library Guide.htm");
}
//---------------------------------------------------------------------------
void __fastcall TBasic4GLv2MainForm::SourceCodeAdvancedPaintGutter(
      TObject *Sender, TCanvas *Canvas, TRect &Rect, int Index)
{

    // Breakpoint icons
    if (m_vm.IsUserBreakPt (Index)) {

        // Display valid breakpoints in full read.
        // Invalid breakpoints are dithered out.
        int image = IsValidBreakPt (Index) ? 1 : 2;
        GutterImages->Draw (Canvas,
                            Rect.Left + 2,
                            (Rect.Top + Rect.Bottom) / 2 - 6,
                            image);
    }

    // Current line icon
    if (m_mode == AP_RUNNING || m_vm.InstructionCount () == 0)
        return;

    int line, col;
    m_vm.GetIPInSourceCode (line, col);
    if (Index == line)
        GutterImages->Draw (Canvas,
                            Rect.Right - 14,
                            (Rect.Top + Rect.Bottom) / 2 - 6,
                            0);
}
//---------------------------------------------------------------------------

bool TBasic4GLv2MainForm::IsValidBreakPt (int line) {
    // Line is valid breakpoint if:
    //  * User has put a breakpoint there, AND either
    //      * There is code that corresponds to the line, OR
    //      * The program has stopped
    //
    // (When the program has stopped, we don't know which breakpoints are valid
    // or not, because the program will be recompiled before it is run.
    // Therefore we display ALL breakpoints as valid.)

    return m_vm.IsUserBreakPt (line)
    && (m_mode == AP_STOPPED || m_vm.GetUserBreakPt (line).m_offset != 0xffff);
}

//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::SourceCodeGetLineAttr(TObject *Sender,
      const AnsiString Line, int Index, const TSelAttrs SelAttrs,
      TLineAttrs &Attrs)
{
    int i = 0;

    // Break line into tokens
    while (i < Line.Length ()) {

        // Skip whitespace
        while (i < Line.Length () && Line [i + 1] <= ' ')
            i++;

        if (i < Line.Length ()) {

            // Identify basic token type
            char c = tolower (Line [i + 1]);
            int left = i;
            int style;
            if ((c >= 'a' && c <= 'z') || c == '_')
                style = 1;                              // 1 = keyword
            else if ((c >= '0' && c <= '9') || c == '.')
                style = 2;                              // 2 = Number
            else if (c == '"')
                style = 3;                              // 3 = String
            else if (c == '\'')
                style = 4;                              // 4 = Comment
            else
                style = 5;                              // 5 = Symbol

            i++;
            bool done = false;
            while (i < Line.Length () && !done) {

                // Look for end of token
                c = tolower (Line [i + 1]);
                switch (style) {
                case 1:
                    if (c == '$' || c == '%' || c== '#') {
                        i++;
                        done = true;
                    }
                    else
                        done = !((c >= 'a' && c <= 'z') || (c >= '0' && c <= '9') || c == '_');
                    break;
                case 2:
                    done = !((c >= '0' && c <= '9') || c == '.');
                    break;
                case 3:
                    done = c == '"';
                    if (done)
                        i++;
                    break;
                case 4:
                    done = false;
                    break;
                case 5:
                    done = true;
                };

                if (!done)
                    i++;
            }

            // Colour in token
            if (i > left) {
                AnsiString token = Line.SubString (left + 1, i - left);         // Get token text
                std::string str = token.c_str ();                               // Convert to STL string
                str = LowerCase (str);

                // Choose text style
                TLineAttr attr;
                switch (style) {
                case 1:
                    if (m_comp.IsReservedWord (str) || m_comp.IsOperator (str))
                        attr = m_keywordStyle;
                    else if (m_comp.IsFunction (str))
                        attr = m_functionStyle;
                    else if (m_comp.IsConstant (str))
                        attr = m_constantStyle;
                    else
                        attr = m_defaultStyle;
                    break;
                case 2:
                    attr = m_numberStyle;
                    break;
                case 3:
                    attr = m_stringStyle;
                    break;
                case 4:
                    attr = m_commentStyle;
                    break;
                case 5:
                    if (m_comp.IsOperator (str))
                        attr = m_keywordStyle;
                    else
                        attr = m_defaultStyle;
                    break;
                };
                for (int x = left; x < i; x++)
                    Attrs [x] = attr;
            }
        }
    }

    // Colour breakpoint lines red
    if (IsValidBreakPt (Index))
        for (int i = 0; i < Line.Length (); i++)
            Attrs [i].BC = m_breakPtLineCol;
}
//---------------------------------------------------------------------------
void TBasic4GLv2MainForm::DefaultStyle (TLineAttr& style) {

    // Reset style to plain window text (used as a starting point)
    style.FC        = SourceCode->Font->Color;
    style.BC        = SourceCode->Color;
    style.Style     = TFontStyles ();
    style.ex_style  = 0;
}
//---------------------------------------------------------------------------
void TBasic4GLv2MainForm::ResetStyles () {

    // Reset syntax highlighting styes
    DefaultStyle (m_defaultStyle);
    DefaultStyle (m_commentStyle);
    DefaultStyle (m_keywordStyle);
    DefaultStyle (m_stringStyle);
    DefaultStyle (m_functionStyle);
    DefaultStyle (m_numberStyle);
    DefaultStyle (m_constantStyle);

    m_commentStyle.FC       = 0x00A77C65;
    m_commentStyle.Style    = TFontStyles () << fsItalic;   // Comments in italic
    m_keywordStyle.FC       = clBlue;                       // Keywords in blue
    m_stringStyle.FC        = clGreen;                      // Strings in green
    m_functionStyle.FC      = clRed;                        // Functions in red
    m_numberStyle.Style     = TFontStyles () << fsBold;     // Numbers are bold
    m_constantStyle.Style   = TFontStyles () << fsBold;     // As are constants
    m_breakPtLineCol        = m_keywordStyle.BC;
    int r = m_breakPtLineCol         & 0xff,
        g = (m_breakPtLineCol >> 8)  & 0xff,
        b = (m_breakPtLineCol >> 16) & 0xff;
    r = 255 - (255 - r) / 4;
    g = g - g / 4;
    b = b - g / 4;
    m_breakPtLineCol = r | g << 8 | b << 16;        
}
//---------------------------------------------------------------------------

// Note: Standard Delphi editing actions don't recognise the TMemoEx component
//       (which we are using as the text editor.)
//       Therefore we manually update and service the actions here.

void __fastcall TBasic4GLv2MainForm::EditCut1Update(TObject *Sender)
{
    ((TAction *) Sender)->Enabled = SourceCode->Focused () && SourceCode->SelLength > 0;
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::EditSelectAll1Update(TObject *Sender)
{
    EditSelectAll1->Enabled = SourceCode->Focused ();
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::EditUndo1Update(TObject *Sender)
{
    EditUndo1->Enabled = SourceCode->CanUndo;
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::EditCut1Execute(TObject *Sender)
{
    SourceCode->CutToClipboard ();
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::EditCopy1Execute(TObject *Sender)
{
    SourceCode->CopyToClipboard ();
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::EditPaste1Execute(TObject *Sender)
{
    SourceCode->PasteFromClipboard ();
}
//---------------------------------------------------------------------------


void __fastcall TBasic4GLv2MainForm::EditSelectAll1Execute(TObject *Sender)
{
    SourceCode->SelectAll ();
}
//---------------------------------------------------------------------------


void __fastcall TBasic4GLv2MainForm::EditUndo1Execute(TObject *Sender)
{
    SourceCode->Undo ();
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::EditDelete1Execute(TObject *Sender)
{
    SourceCode->DeleteSelected ();    
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::FindActionUpdate(TObject *Sender)
{
    ((TAction *) Sender)->Enabled = SourceCode->Focused ();
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::FindNextActionUpdate(TObject *Sender)
{
    FindNextAction->Enabled = SourceCode->Focused () && FindDialog->FindText != "";
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::FindActionExecute(TObject *Sender)
{
    FindDialog->Execute ();
}
//---------------------------------------------------------------------------

bool TBasic4GLv2MainForm::Find (AnsiString text, bool caseSensitive) {

    if (!caseSensitive)
        text = text.LowerCase ();
        
    int sx = SourceCode->CaretX + 1, sy = SourceCode->CaretY;       // Start position
    int x = sx, y = sy;                                             // Current position
    bool startLine = true;
                                           
    // Search through lines
    while (y != sy || startLine) {

        // Extract line
        AnsiString line = SourceCode->Lines->operator[] (y);
        if (!caseSensitive)
            line = line.LowerCase ();

        // Search for string after x
        int p = line.SubString (x, 9999).Pos (text);
        if (p != 0) {
            p += x - 1;
            SourceCode->SelStart = SourceCode->PosFromCaret (p, y);
            SourceCode->SelLength = text.Length ();
            return true;
        }
        if (++y >= SourceCode->Lines->Count)
            y = 0;
        x = 1;
        startLine = false;
    }
    return false;
}
//---------------------------------------------------------------------------

bool TBasic4GLv2MainForm::Replace (AnsiString findText, AnsiString replaceText, bool caseSensitive) {
    while (Find (findText, caseSensitive))
        switch (MessageDlg (    "Replace this occurance of \"" + findText + "\"?",
                mtConfirmation,
                TMsgDlgButtons() << mbYes << mbNo << mbCancel,
                0)) {
            case mrYes:     SourceCode->SelText = replaceText;
                            break;
            case mrCancel:  return false;
            case mrNo:      break;
        }
    return false;
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::PlayPauseActionExecute(
      TObject *Sender)
{
    switch (m_mode) {
    case AP_RUNNING:
        m_vm.Pause ();                  // Pause program
        break;

    case AP_STOPPED:
        if (Compile ())                 // When stopped, Play is exactly the same as Run
            Continue ();
        break;

    case AP_PAUSED:
        Continue ();                    // When paused, play continues from where program was halted.
        break;
    };
}
//---------------------------------------------------------------------------
void TBasic4GLv2MainForm::ToggleBreakPt (int line) {

    // Place or clear breakpoint on line
    if (m_vm.IsUserBreakPt (line))  m_vm.ClearUserBreakPt (line);
    else                            m_vm.SetUserBreakPt (line, vmUserBreakPt ());

    // Repaint source code (to update the gutter, and highlight the line)
    SourceCode->InvalidateLine (line);
    SourceCode->Invalidate ();
}

//---------------------------------------------------------------------------
void __fastcall TBasic4GLv2MainForm::SourceCodeGutterClick(TObject *Sender,
      int X, int Y, int Index)
{
    // Clicking in the gutter toggles a breakpoint
    ToggleBreakPt (Index);
}
//---------------------------------------------------------------------------

void TBasic4GLv2MainForm::DoStep (int type) {
    if (m_mode == AP_RUNNING)
        return;

    // Recompile program if necessary
    if (m_mode == AP_STOPPED && !Compile ())
        return;

    // Patch in temp breakpoints
    switch (type) {
    case 1: m_vm.AddStepBreakPts (false); break;        // Step over
    case 2: m_vm.AddStepBreakPts (true);  break;        // Step into
    case 3:
        if (!m_vm.AddStepOutBreakPt ())                 // Step out
            return;                                     // (No gosub to step out of)
        break;
    }

    // Resume running program
    m_delayScreenSwitch = true;
    Continue ();
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::StepActionExecute(TObject *Sender)
{
    DoStep (1);    
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::StepIntoActionExecute(TObject *Sender)
{
    DoStep (2);    
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::StepOutOfActionExecute(
      TObject *Sender)
{
    DoStep (3);
}
//---------------------------------------------------------------------------

void TBasic4GLv2MainForm::RefreshDebugDisplays () {

    // Show/hide debug controls
    DebugBar->Visible       = m_debugMode;
    DebugPanel->Visible     = m_debugMode;
    DebugSplitter->Visible  = m_debugMode;
    DebugMenu->Visible      = m_debugMode;

    PlayPauseAction->ImageIndex = m_mode == AP_RUNNING ? 9 : 8;
    PlayPauseAction->Enabled    = true;
    StepAction->Enabled         = m_mode != AP_RUNNING;
    StepIntoAction->Enabled     = m_mode != AP_RUNNING;
    StepOutOfAction->Enabled    = m_mode == AP_PAUSED && m_vm.CallStack ().size () > 0;

    if (!m_debugMode)
        return;

    // Clear debug controls
    WatchListBox->Clear ();
    GosubListBox->Clear ();

    for (   std::list<std::string>::iterator i = m_watches.begin ();
            i != m_watches.end ();
            i++)
        WatchListBox->Items->Add (((*i) + ": " + EvaluateWatch ((*i), true)).c_str ());
    WatchListBox->Items->Add ("");              // Last line is blank, and can be clicked on to add new watch

    if (m_mode != AP_PAUSED)
        return;

    // Update call stack
    GosubListBox->Items->Add ("IP");
    for (int i2 = m_vm.CallStack ().size () - 1; i2 >= 0; i2--)
        GosubListBox->Items->Add (((std::string) "gosub " + m_comp.DescribeStackCall (m_vm.CallStack () [i2])).c_str ());
    GosubListBox->ItemIndex = 0;        
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::DebugActionExecute(TObject *Sender)
{
    // Toggle debug mode
    m_debugMode = !m_debugMode;
    DebugAction->Checked = m_debugMode;
    RefreshDebugDisplays ();
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::GosubListBoxClick(TObject *Sender)
{
    // Top most element is special, and corresponds to the current IP
    if (GosubListBox->ItemIndex == 0) {
        int line, col;
        m_vm.GetIPInSourceCode (line, col);
        PlaceCursor (line, col);
    }
    else {

        // Find stack entry clicked on
        int index = m_vm.CallStack ().size () - GosubListBox->ItemIndex;
        if (index >= 0 && index < m_vm.CallStack ().size ()) {

            // Find return address
            // Subtract 1 to get gosub instruction address
            unsigned int offset = m_vm.CallStack () [index] - 1;
            if (offset < m_vm.InstructionCount ()) {

                // Lookup op-code and extract source code position
                int row = m_vm.Instruction (offset).m_sourceLine,
                    col = m_vm.Instruction (offset).m_sourceChar;

                // Position cursor
                PlaceCursor (row, col);
            }
        }
    }
}
//---------------------------------------------------------------------------

std::string TBasic4GLv2MainForm::EvaluateWatch (std::string watch, bool canCallFunc) {
    if (m_mode == AP_RUNNING)
        return "???";

    // Save virtual machine state
    vmState state   = m_vm.GetState ();
    ApMode saveMode = m_mode;
    try {

        // Compile watch expression
        // This also gives us the expression result type
        unsigned int codeStart = m_vm.InstructionCount ();
        vmValType valType;
        if (!m_comp.TempCompileExpression (watch, valType))
            return m_comp.GetError ();

        if (!canCallFunc)
            // Expressions aren't allowed to call functions for mouse-over hints.
            // Scan compiled code for OP_CALL_FUNC or OP_CALL_OPERATOR_FUNC
            for (int i = codeStart; i < m_vm.InstructionCount (); i++)
                if (m_vm.Instruction (i).m_opCode == OP_CALL_FUNC
                ||  m_vm.Instruction (i).m_opCode == OP_CALL_OPERATOR_FUNC)
                    return "Mouse hints can't call functions. Use watch instead.";

        // Run compiled code
        m_vm.GotoInstruction (codeStart);
        m_mode = AP_RUNNING;
        try {
            do {

                // Run the virtual machine for a certain number of steps
                m_vm.Continue (gbStepsUntilRefresh);

                // Process windows messages (to keep application responsive)
                Application->ProcessMessages ();
                m_glWin->ProcessWindowsMessages();
                if (m_glWin->PausePressed ())           // Check for pause key. (This allows us to pause when in full screen mode. Useful for debugging.)
                    m_vm.Pause ();
            } while (       m_mode == AP_RUNNING
                        &&  !m_vm.Error ()
                        &&  !m_vm.Done ()
                        &&  !m_vm.Paused ()
                        &&  !m_glWin->Closing ());
        }
        catch (...) {
            // Basic4GL is designed not to throw exceptions.
            // But just to be sure...
            m_vm.MiscError ("An exception occured");
        }

        // Error occurred?
        if (m_vm.Error ())
            return m_vm.GetError ();

        // Execution didn't finish?
        if (!m_vm.Done ())
            return "???";

        // Convert expression result to string
        if (valType == VTP_STRING)                                  // String is special case.
            return "\"" + m_vm.RegString () + "\"";                 // Stored in string register.
        else {
            std::string temp = m_vm.ValToString (m_vm.Reg (), valType, VM_DATATOSTRINGMAXCHARS);
            return temp;
        }
//            return m_vm.ValToString (m_vm.Reg (), valType, VM_DATATOSTRINGMAXCHARS);   // Otherwise standard vmValue, with type info
    }
    __finally {
        m_vm.SetState (state);
        m_mode = saveMode;
        RefreshActions ();
    }
    return "";               // (Should never get here.)
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::WatchListBoxDblClick(TObject *Sender)
{
    EditWatch ();    
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::WatchListBoxKeyPress(TObject *Sender,
      char &Key)
{
    if (Key == 13)
        EditWatch ();
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::WatchListBoxKeyUp(TObject *Sender,
      WORD &Key, TShiftState Shift)
{
    if (Key == VK_DELETE)
        DeleteWatch ();
    else if (Key == VK_INSERT) {
        WatchListBox->ItemIndex = m_watches.size ();
        EditWatch ();
    }
}
//---------------------------------------------------------------------------
void TBasic4GLv2MainForm::EditWatch () {
    AnsiString newWatch, oldWatch;

    // Find watch
    int index = WatchListBox->ItemIndex;
    int saveIndex = index;
    std::list<std::string>::iterator i = m_watches.begin ();
    while (index > 0) {
        i++;
        index--;
    }

    // Extract watch text
    if (i != m_watches.end ())  oldWatch = (*i).c_str ();
    else                        oldWatch = "";

    // Prompt for new text
    newWatch = InputBox ("Watch variable", "Enter variable/expression:", oldWatch);

    // Update/insert/delete watch
    if (newWatch == "") {
        if (i != m_watches.end ())
            m_watches.erase (i);
    }
    else {
        if (i != m_watches.end ())
            *i = newWatch.c_str ();
        else
            m_watches.push_back (newWatch.c_str ());
    }

    RefreshDebugDisplays ();
    WatchListBox->ItemIndex = saveIndex;
    UpdateWatchHint ();
}
//---------------------------------------------------------------------------

void TBasic4GLv2MainForm::DeleteWatch () {

    // Find watch
    int index = WatchListBox->ItemIndex;
    int saveIndex = index;
    std::list<std::string>::iterator i = m_watches.begin ();
    while (index > 0) {
        i++;
        index--;
    }

    // Delete watch
    if (i != m_watches.end ())
        m_watches.erase (i);

    RefreshDebugDisplays ();
    WatchListBox->ItemIndex = saveIndex;
    UpdateWatchHint ();
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::WatchListBoxClick(TObject *Sender)
{
    UpdateWatchHint ();
}
//---------------------------------------------------------------------------

void TBasic4GLv2MainForm::UpdateWatchHint () {
    int index = WatchListBox->ItemIndex;
    if (index >= 0 && index < WatchListBox->Items->Count - 1)
        WatchListBox->Hint = WatchListBox->Items->Strings [index];
    else
        WatchListBox->Hint = "";
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::AddWatchActionExecute(TObject *Sender)
{
    // Switch to debug mode
    m_debugMode = true;
    DebugAction->Checked = true;

    // If text is selected, add it as a watch.
    // Otherwise determine variable based on cursor position.
    if (SourceCode->SelLength > 0)
        m_watches.push_back (SourceCode->SelText.c_str ());
    else
        m_watches.push_back (GetVariableAt (SourceCode->CaretX, SourceCode->CaretY));

    // Refresh debug display (to display new watch)
    RefreshDebugDisplays ();
    UpdateWatchHint ();
}
//---------------------------------------------------------------------------


void __fastcall TBasic4GLv2MainForm::SourceCodeMouseMove(TObject *Sender,
      TShiftState Shift, int X, int Y)
{
    if (m_debugMode && m_mode != AP_RUNNING && X >= SourceCode->GutterWidth) {

        // Find character mouse is pointing to
        int cx, cy;
        SourceCode->Mouse2Caret (X, Y, cx, cy);
        // (Note: I had to modify the MemoEx source to make Mouse2Caret public..)

        // Special case: If text is selected, and selection is all in the same
        // line, AND mouse is pointing inside selection, then evaluate selection
        // text.
        std::string variable = "";
        if (SourceCode->SelLength > 0) {

            // Find start and end positions
            int sx, sy, ex, ey;
            SourceCode->CaretFromPos(SourceCode->SelStart, sx, sy);
            SourceCode->CaretFromPos(SourceCode->SelStart + SourceCode->SelLength - 1, ex, ey);

            if (    sy == ey                // Start and end on same line
                &&  sy == cy                // Mouse pointing to same line
                &&  cx >= sx && cx < ex)    // Mouse pointing inside selection
                variable = SourceCode->SelText.c_str ();
        }

        // Otherwise scan text under pointer to find appropriate variable name
        if (variable == "")
            variable = GetVariableAt (cx, cy);

        // Show variable and value as popup hint
        if (variable != "") {
            SourceCode->Hint = (variable + ": " + EvaluateWatch (variable, false)).c_str ();
            SourceCode->ShowHint = true;
            return;
        }
    }
    SourceCode->Hint = "";
    SourceCode->ShowHint = false;
}
//---------------------------------------------------------------------------

std::string TBasic4GLv2MainForm::GetVariableAt (int x, int y) {

    // Find line
    if (y < 0 || y >= SourceCode->Lines->Count)
        return "";
    std::string line = SourceCode->Lines->operator[] (y).c_str ();

    // Find character
    if (x < 0 || x >= line.length () || line [x] <= ' ')
        return "";

    // Scan to right of word
    int right = x + 1;
    while (right < line.length ()
        && (    (line [right] >= 'a' && line [right] <= 'z')
            ||  (line [right] >= 'A' && line [right] <= 'Z')
            ||  (line [right] >= '0' && line [right] <= '9')
            ||  line [right] == '_'
            ||  line [right] == '#'
            ||  line [right] == '$'))
        right++;

    // Scan left
    int left = x;
    while (left >= 0
        && (    (line [left] >= 'a' && line [left] <= 'z')
            ||  (line [left] >= 'A' && line [left] <= 'Z')
            ||  (line [left] >= '0' && line [left] <= '9')
            ||  line [left] == '.'
            ||  line [left] == '_'
            ||  line [left] == '#'
            ||  line [left] == '$'
            ||  line [left] == ')')) {

        // Skip over brackets
        if (line [left] == ')') {
            int level = 1;
            left--;
            while (level > 0 && left >= 0) {
                if (line [left] == ')')         level++;
                else if (line [left] == '(')    level--;
                left--;
            }
            while (left >= 0 && line [left] <= ' ')
                left--;
        }
        else
            left--;
    }
    left++;

    // Trim whitespace from left
    while (left < right && line [left] <= ' ')
        left++;

    // Return result
    if (left < right)
        return line.substr (left, right - left);
    else
        return "";
};
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::SymbolicDebuggerGuide1Click(
      TObject *Sender)
{
    ShowDoc ("Symbolic Debugger Guide.htm");
}
//---------------------------------------------------------------------------


void __fastcall TBasic4GLv2MainForm::FindDialogFind(TObject *Sender)
{
    FindDialog->CloseDialog ();
    FindNextActionExecute (this);
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::FindNextActionExecute(TObject *Sender)
{
    if (FindDialog->FindText != "")
        if (!Find (FindDialog->FindText, FindDialog->Options.Contains(frMatchCase)))
            MessageDlg (    "\"" + FindDialog->FindText + "\" not found.",
                            mtInformation,
                            TMsgDlgButtons() << mbOK,
                            0);
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::ReplaceActionExecute(TObject *Sender)
{
    ReplaceDialog->Execute ();    
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::ReplaceDialogReplace(TObject *Sender)
{
    ReplaceDialog->CloseDialog ();
    Replace  (ReplaceDialog->FindText, ReplaceDialog->ReplaceText, ReplaceDialog->Options.Contains (frMatchCase));
}
//---------------------------------------------------------------------------

bool TBasic4GLv2MainForm::RunAndQuit (std::string filename) {
    if (filename == "")
        return false;

    try {

        // Remove quotes (if any
        if (filename [0] == '\"')
            filename = filename.substr(1, filename.length () - 1);
        if (filename [filename.length () - 1] == '\"')
            filename = filename.substr(0, filename.length () - 1);

        // Find full path and filename
        char buffer [1024], *nameStart;
        GetFullPathName (filename.c_str (), 1024, buffer, &nameStart);
        std::string fileBit = nameStart;
        *nameStart = 0;
        std::string pathBit = buffer;

        // Set current directory to file
        GetCurrentDirectory (1024, buffer);                 // Save previous directory
        try {
            SetCurrentDirectory (pathBit.c_str ());             // Set new one

            // Read file into compiler
            TFileStream *file = new TFileStream (fileBit.c_str (), fmOpenRead);
            try {
                m_comp.Parser ().SourceCode ().clear ();
                while (file->Position < file->Size)
                    m_comp.Parser ().SourceCode().push_back(ReadString (*file).c_str ());
            }
            __finally {
                delete file;
            }

            // Compile it
            m_comp.ClearError ();
            m_comp.Compile ();
            if (m_comp.Error ())
                throw Exception (m_comp.GetError ().c_str ());

            // Reset Virtual machine
            m_vm.Reset ();

            // Reset OpenGL state
            m_glWin->Activate ();
//            m_glWin->InitGL ();

            // Run it
            try {
                do {
                    m_vm.Continue (gbStepsUntilRefresh);
                    Application->ProcessMessages ();
                    m_glWin->ProcessWindowsMessages();
                } while (       !m_vm.Error ()
                            &&  !m_vm.Done ()
                            &&  !m_glWin->Closing ());
            }
            catch (...) {
                // Basic4GL is designed not to throw exceptions.
                // But just to be sure...
                m_vm.MiscError ("An exception occured");
            }
            if (m_vm.Error ()) {
                if (m_glWin->FullScreen ())
                    m_glWin->Hide ();
                throw Exception (m_vm.GetError ().c_str ());
            }

            // If program terminated normally, then show screen until keypress
            // (fullscreen) or window is closed (windowed)

            if (m_glWin->Visible () && m_vm.Done ()) {    // Program ended normally
                if (m_glWin->FullScreen ()) {

                    // Wait for keypress.
                    m_glWin->ProcessWindowsMessages();
                    while (m_glWin->GetKey () != 0) m_glWin->ProcessWindowsMessages();
                    while (m_glWin->GetKey () == 0) m_glWin->ProcessWindowsMessages();
                }
                else {
                    while (!m_glWin->Closing ())
                        m_glWin->ProcessWindowsMessages ();
                }
            }

            // Hide OpenGL window
            m_glWin->Hide ();
        }
        __finally {
            SetCurrentDirectory (buffer);                 // Restore previous directory
        }
    }
    catch (Exception &exception)
    {
         Application->ShowException(&exception);
    }
    catch (...)
    {
         try
         {
             throw Exception("");
         }
         catch (Exception &exception)
         {
             Application->ShowException(&exception);
         }
    }
    return true;
}

//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::CreateStandAloneExecute(
      TObject *Sender)
{
    // Create standalone executable
    // Run "create standalone" form
    LoadProgramIntoCompiler ();
    TCreateStandaloneForm *dlg = new TCreateStandaloneForm (this);
    try {
        dlg->Execute (m_comp, ExtractFileName (m_filename));
    }
    __finally {
        dlg->Release ();
    }
}
//---------------------------------------------------------------------------

void __fastcall TBasic4GLv2MainForm::CreatingStandaloneexes1Click(
      TObject *Sender)
{
    ShowDoc ("Creating Standalone exes.htm");
}
//---------------------------------------------------------------------------

void TBasic4GLv2MainForm::LoadScreenRegistrySettings (int& width, int& height,
    int& bpp, bool& fullScreen, bool& border, ResetGLModeType& resetMode) {

    // Default settings
    int screenMode = 1, colDepth = 1;
    fullScreen = true;
    border = true;

    // Read from registry
    try {
        TRegistry *Reg = new TRegistry ();
        try
        {
            Reg->RootKey = HKEY_CURRENT_USER;
            if (Reg->OpenKeyReadOnly("\\Software\\Tom\\Basic4GLv2")) {
                screenMode  = Reg->ReadInteger ("ScreenMode");
                colDepth    = Reg->ReadInteger ("ColourDepth");
                fullScreen  = Reg->ReadBool    ("FullScreen");
                border      = Reg->ReadBool    ("Border");
                resetMode   = Reg->ReadInteger ("ResetGLMode");
            }
        }
        __finally
        {
            delete Reg;
        }
    } catch (...) { ; }

    // Process into width, height and bpp
    width = Screen->Width;
    height = Screen->Height;
    switch (screenMode) {
    case 1: width   = 640;              height  = 480;              break;
    case 2: width   = 800;              height  = 600;              break;
    case 3: width   = 1024;             height  = 768;              break;
    case 4: width   = 1280;             height  = 1024;             break;
    }
    bpp = 0;
    switch (colDepth) {
    case 1: bpp = 16; break;
    case 2: bpp = 32; break;
    }
}
//---------------------------------------------------------------------------

