//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "CreateStandalone.h"
#include "resource.h"
#include "Standalone.h"
#include "Basic4GLSettings.h"
#include <Registry.hpp>
#include <windows.h>
#include <winnt.h>
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma link "ScreenMode"
#pragma resource "*.dfm"
TCreateStandaloneForm *CreateStandaloneForm;

inline void Error (AnsiString desc) {
    MessageDlg (desc, mtError, TMsgDlgButtons () << mbOK, 0);
}

//---------------------------------------------------------------------------
__fastcall TCreateStandaloneForm::TCreateStandaloneForm(TComponent* Owner)
    : TForm(Owner)
{
}
//---------------------------------------------------------------------------
AnsiString ApplicationPath () {
    return IncludeTrailingPathDelimiter (ExtractFilePath (Application->ExeName));
}

void TCreateStandaloneForm::Execute (TomBasicCompiler& comp, AnsiString filename) {

    if (Trim (filename) == "")
        filename = "MyProgram.gb";

    // Ensure the program compiles.
    comp.ClearError ();
    comp.Compile ();
    if (comp.Error ()) {
        Error ("Program does not compile.\r\nPlease fix the problem then try again.");
        return;
    }

    // Must produce some code
    if (comp.VM ().InstructionCount () <= 1) {
        Error ("Program does not produce any executable code.\r\nPlease fix the problem then try again.");
        return;
    }

    // Initialise form and controls
    LoadFromRegistry ();
    Title->Text     = ChangeFileExt (filename, "");
    FileName->Text  = ChangeFileExt (IncludeTrailingPathDelimiter (GetCurrentDir ()) + filename, ".exe");

    // Look for filenames in the constants strings.
    Files->Clear ();
    int i;
    for (i = 0; comp.VM().StringConstants().IndexValid(i); i++)
        AddFile (comp.VM().StringConstants().Value(i).c_str ());

    bool done = false;
    while (!done) {
        if (ShowModal () == mrOk) {

            // Check whether file exists
            if (    !FileExists (FileName->Text)
                ||  MessageDlg("File exists! Overwrite?", mtConfirmation, TMsgDlgButtons() << mbYes << mbCancel, 0) == mrYes) {
                CreateStandAlone (comp);
                done = true;
            }
        }
        else
            done = true;
    }
    
    SaveToRegistry ();
}
//---------------------------------------------------------------------------


void __fastcall TCreateStandaloneForm::AddEmbeddedClick(TObject *Sender)
{
    if (OpenDialog->Execute ())
        for (int i = 0; i < OpenDialog->Files->Count; i++)
            AddFile (OpenDialog->Files->operator[] (i));
}
//---------------------------------------------------------------------------
void TCreateStandaloneForm::AddFile (AnsiString name) {
    // Check that the string is a filename
    if (FileExists (name)) {

        // Process filename for comparisons
        AnsiString compFileName = ProcessPath (name.c_str ()).c_str ();

        // Must not be already present
        if (Files->Items->IndexOf (compFileName) < 0)
            Files->Items->Add (compFileName);           // Add it to list
    }
}
//---------------------------------------------------------------------------

void __fastcall TCreateStandaloneForm::DeleteEmbeddedClick(TObject *Sender)
{
    Files->DeleteSelected ();
}
//---------------------------------------------------------------------------

void __fastcall TCreateStandaloneForm::UPXBrowseClick(TObject *Sender)
{
    OpenUPX->FileName = UPXFile->Text;
    OpenUPX->InitialDir = ExtractFilePath (UPXFile->Text);
    if (OpenUPX->Execute ())
        UPXFile->Text = OpenUPX->FileName;
}
//---------------------------------------------------------------------------

void __fastcall TCreateStandaloneForm::UseUPXClick(TObject *Sender)
{
    UPXFile->Enabled = UseUPX->Checked;
    UPXBrowse->Enabled = UseUPX->Checked;
}
//---------------------------------------------------------------------------

void __fastcall TCreateStandaloneForm::FileNameBrowseClick(TObject *Sender)
{
    SaveDialog->FileName = FileName->Text;
    SaveDialog->InitialDir = ExtractFilePath (FileName->Text);
    if (SaveDialog->Execute ())
        FileName->Text = SaveDialog->FileName;
}
//---------------------------------------------------------------------------
void TCreateStandaloneForm::LoadFromRegistry () {

    // Use current screen mode settings if none are defined for standalones
    ScreenModeFrame->LoadFromRegistry ();

    // Load settings from registry
    try {
        TRegistry *Reg = new TRegistry ();
        try
        {
            Reg->RootKey = HKEY_CURRENT_USER;
            if (Reg->OpenKeyReadOnly(REG_KEY)) {
                FinishedOption->ItemIndex           = Reg->ReadInteger  ("StandAlone_Finished");
                ErrorOption->ItemIndex              = Reg->ReadInteger  ("StandAlone_Error");
                UseUPX->Checked                     = Reg->ReadBool     ("StandAlone_UseUPX");
                UPXFile->Text                       = Reg->ReadString   ("StandAlone_UPXFile");
                ScreenModeFrame->ScreenMode->ItemIndex   = Reg->ReadInteger  ("StandAlone_ScreenMode");
                ScreenModeFrame->ColourDepth->ItemIndex  = Reg->ReadInteger  ("StandAlone_ColourDepth");
                ScreenModeFrame->FullScreen->Checked     = Reg->ReadBool     ("StandAlone_FullScreen");
                EscKeyQuits->Checked                = Reg->ReadBool     ("StandAlone_EscKeyQuits");
                ScreenModeFrame->Border->Checked        = Reg->ReadBool ("Standalone_Border");
            }
        }
        __finally
        {
            delete Reg;
        }
    } catch (...) {;}
}
//---------------------------------------------------------------------------
void TCreateStandaloneForm::SaveToRegistry () {

    // Save settings from registry
    TRegistry *Reg = new TRegistry;
    try
    {
        Reg->RootKey = HKEY_CURRENT_USER;
        if (Reg->OpenKey(REG_KEY, true)) {
            Reg->WriteInteger   ("StandAlone_Finished",     FinishedOption->ItemIndex);
            Reg->WriteInteger   ("StandAlone_Error",        ErrorOption->ItemIndex);
            Reg->WriteBool      ("StandAlone_UseUPX",       UseUPX->Checked);
            Reg->WriteString    ("StandAlone_UPXFile",      UPXFile->Text);
            Reg->WriteInteger   ("StandAlone_ScreenMode",   ScreenModeFrame->ScreenMode->ItemIndex );
            Reg->WriteInteger   ("StandAlone_ColourDepth",  ScreenModeFrame->ColourDepth->ItemIndex);
            Reg->WriteBool      ("StandAlone_FullScreen",   ScreenModeFrame->FullScreen->Checked   );
            Reg->WriteBool      ("StandAlone_EscKeyQuits",  EscKeyQuits->Checked);
            Reg->WriteBool      ("StandAlone_Border",       ScreenModeFrame->Border->Checked       );
        }
    }
    __finally
    {
        delete Reg;
    }
}
//---------------------------------------------------------------------------
#define COPYBLOCKSIZE   0x10000

void CopyFileBytes (GenericIStream& src, GenericOStream& dst, int count) {

    // Copy data in blocks.
    char data [COPYBLOCKSIZE];
    while (count >= COPYBLOCKSIZE) {
        src.read (data, COPYBLOCKSIZE);
        dst.write (data, COPYBLOCKSIZE);
        count -= COPYBLOCKSIZE;
    }
    if (count > 0) {
        src.read (data, count);
        dst.write (data, count);
    }
}

void PadFile (GenericOStream& dst, int count) {

    // Pad out zeros in blocks.
    char data [COPYBLOCKSIZE];
    memset (data, 0, COPYBLOCKSIZE);
    while (count >= COPYBLOCKSIZE) {
        dst.write (data, COPYBLOCKSIZE);
        count -= COPYBLOCKSIZE;
    }
    if (count > 0)
        dst.write (data, count);
}

int PadLength (int length, int pageSize) {
    int remainder = length % pageSize;
    return remainder == 0 ? length : length + pageSize - remainder;
}

//---------------------------------------------------------------------------
void TCreateStandaloneForm::CreateStandAlone (TomBasicCompiler& comp) {

    // Setup patch data.
    // Write it to a stream first, so that we can calculate the size
    std::stringstream patchData;

    // Header first
    StandaloneHeader header;
    header.version = STANDALONE_VERSION;
    memset (header.title, 0, 32);
    assert (Title->Text.Length () < 32);
    memcpy (header.title, Title->Text.c_str (), Title->Text.Length ());
    header.finishedOption   = FinishedOption->ItemIndex;
    header.errorOption      = ErrorOption->ItemIndex;
    header.screenMode       = ScreenModeFrame->ScreenMode->ItemIndex;
    header.colourDepth      = ScreenModeFrame->ColourDepth->ItemIndex;
    header.fullScreen       = ScreenModeFrame->FullScreen->Checked;
    header.escKeyQuits      = EscKeyQuits->Checked;
    header.border           = ScreenModeFrame->Border->Checked;
    patchData.write ((char *) &header, sizeof (header));

    // Next comes program
    std::string program = "";
    StringVector& sourceCode = comp.Parser ().SourceCode ();
    for (   StringVector::iterator i = sourceCode.begin ();
            i != sourceCode.end ();
            i++)
        program += (*i) + "\r";
    int size = program.length () + 1;
    patchData.write ((char *) &size, sizeof (size));
    patchData.write (program.c_str (), size);

    // Then embedded files
    int fileCount = Files->Items->Count + 1;
        // We always embed charset.png (although this may be made optional in a
        // later version) so that the program can display text.
    patchData.write ((char *) &fileCount, sizeof (fileCount));

    // We need to do a bit of current directory swapping to make charset.png
    // appear as if it's in the same directory as the program.
    char filePath [1024];
    GetCurrentDirectory (1024, filePath);
    SetCurrentDirectory (ExtractFilePath (Application->ExeName).c_str ());
    if (!EmbedFile ("charset.png", patchData)) {
        SetCurrentDirectory (filePath);
        Error ("Couldn't embed charset.png");
        return;
    }
    SetCurrentDirectory (filePath);

    // Embedd regular files
    for (int j = 0; j < Files->Items->Count; j++) {
        AnsiString filename = Files->Items->Strings [j];
        if (!EmbedFile (filename.c_str (), patchData)) {
            Error ("Couldn't embed " + filename);
            return;
        }
    }

    // Calculate new data length, and length of data section (padded length)
    int dataLength          = patchData.tellp ();
    int paddedDataLength    = PadLength (dataLength, 0x1000);

    // Scan source file.
    // Extract information and verify file integrity.
    //
    // Info:
    //  * Current size of Basic4GL section
    //
    // Integrity:
    //  * File is windows executable
    //  * Last section is Basic4GL section

    // Open file
    AnsiString stub = IncludeTrailingPathDelimiter (ExtractFilePath (Application->ExeName)) + "Stub.bin";
    std::ifstream src (stub.c_str (), std::ios::in | std::ios::binary);
    if (src.fail ()) {
        Error ("Error opening input file");
        return;
    }

    // Read MSDOS header
	IMAGE_DOS_HEADER msdos;
	memset (&msdos, 0, sizeof (msdos));
	src.read ((char *) &msdos, sizeof (msdos));
	if (msdos.e_magic != IMAGE_DOS_SIGNATURE) {
		Error ("Invalid MS_DOS signature. Is this really an exe file?");
		return;
	}

	// Read NT header
	IMAGE_NT_HEADERS32 nt;
	memset (&nt, 0, sizeof (nt));
	src.seekg (msdos.e_lfanew);
	src.read ((char *) &nt, sizeof (nt));
	if (nt.Signature != IMAGE_NT_SIGNATURE) {
		Error ("Invalid NT signature. Is this really a windows exe file?");
		return;
	}

    // Section headers start immediately after nt header.
    // Last section must be named Basic4GL.
    src.seekg ((nt.FileHeader.NumberOfSections - 1) * sizeof (IMAGE_SECTION_HEADER), std::ios::cur);
    IMAGE_SECTION_HEADER section;
    src.read ((char *) &section, sizeof (section));
    if (memcmp (section.Name, "Basic4GL", 8) != 0) {
        Error ("Basic4GL section not found. Is this really a Basic4GL stub?");
        return;
    }

    // Now have all the information we need to start copying and patching the
    // file.

    // Open output file
    std::ofstream dst (FileName->Text.c_str (), std::ios::out | std::ios::binary);
    if (dst.fail ()) {
        Error ("Error opening output file");
        return;
    }

    // Copy up to start of NT header
    src.seekg (0, std::ios::beg);
    CopyFileBytes (src, dst, msdos.e_lfanew);

    // Update NT header to reflect patched file
    nt.OptionalHeader.SizeOfInitializedData += paddedDataLength - section.SizeOfRawData;
    nt.OptionalHeader.SizeOfImage += paddedDataLength - section.SizeOfRawData;

    // Write new header
    dst.write ((char *) &nt, sizeof (nt));
    src.seekg (sizeof (nt), std::ios::cur);

    // Copy regular section headers
    CopyFileBytes (src, dst, (nt.FileHeader.NumberOfSections - 1) * sizeof (IMAGE_SECTION_HEADER));

    // Update Basic4GL section header
    section.Misc.VirtualSize    = dataLength;           // (Virtual size does not include padding)
    section.SizeOfRawData       = paddedDataLength;

    // Write Basic4GL section header
    dst.write ((char *) &section, sizeof (section));
    src.seekg (sizeof (section), std::ios::cur);

    // Copy data upto start of Basic4GL section
    CopyFileBytes (src, dst, section.PointerToRawData - src.tellg ());

    // Write Basic4GL section
    patchData.seekg (0, std::ios::beg);
    CopyStream (patchData, dst, dataLength);
    PadFile (dst, paddedDataLength - dataLength);

    src.close ();
    dst.close ();

    // Invoke UPX
    if (UseUPX->Checked) {
        ShellExecute (  NULL,
                        "open",
                        UPXFile->Text.c_str (),
                        ExtractFileName (FileName->Text).c_str (),
                        ExtractFilePath (FileName->Text).c_str (),
                        SW_SHOW);
    }

    // All done!
    AnsiString  folder  = ExtractFilePath (FileName->Text),
                file    = ExtractFileName (FileName->Text);
/*    ShellExecute (  NULL,
                    "open",
                    folder.c_str (),
                    "",
                    folder.c_str (),
                    SW_SHOW);*/

    MessageDlg (file + " created", mtInformation, TMsgDlgButtons () << mbOK, 0);
}
//---------------------------------------------------------------------------

