//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "VirtualMachineView.h"
//---------------------------------------------------------------------------
#ifndef _MSC_VER
#pragma package(smart_init)                               
#endif
#pragma resource "*.dfm"
TVirtualMachineViewForm *VirtualMachineViewForm;
//---------------------------------------------------------------------------
__fastcall TVirtualMachineViewForm::TVirtualMachineViewForm(TComponent* Owner, TomBasicCompiler& comp)
    : m_vm (comp.VM ()), m_comp (comp), TForm(Owner)
{
    m_updating = false;
}
//---------------------------------------------------------------------------

void SetRowCount (TStringGrid *grid, int count) {
    if (count <= 0) {
        grid->RowCount = 2;
        for (int i = 0; i < grid->ColCount; i++)
            grid->Cells [i] [1] = "";
    }
    else
        grid->RowCount = count + 1;
    grid->Tag = count;
}

void TVirtualMachineViewForm::UpdateData () {

    m_updating = true;
    
    SetRowCount (CodeGrid, 0);
    CodeGrid->Cells [0][0] = "";
    CodeGrid->Cells [1][0] = "Op-code";
    CodeGrid->Cells [2][0] = "Data type";
    CodeGrid->Cells [3][0] = "Value";
    SetRowCount (StrucGrid, 0);
    StrucGrid->Cells [0][0] = "Name";
    StrucGrid->Cells [1][0] = "Data size";
    StrucGrid->Cells [2][0] = "Strings?";
    StrucGrid->Cells [3][0] = "Arrays?";
    SetRowCount (FieldGrid, 0);
    FieldGrid->Cells [0][0] = "Name";
    FieldGrid->Cells [1][0] = "Data size";
    SetRowCount (VarGrid, 0);
    VarGrid->Cells [0][0] = "Name";
    VarGrid->Cells [1][0] = "Data";
    SetRowCount (MemGrid, 0);
    MemGrid->Cells [0][0] = "";
    MemGrid->Cells [1][0] = "Int";
    MemGrid->Cells [2][0] = "Real";
    MemGrid->Cells [3][0] = "String";
    SetRowCount (StackGrid, 0);
    StackGrid->Cells [0][0] = "";
    StackGrid->Cells [1][0] = "Int";
    StackGrid->Cells [2][0] = "Real";
    StackGrid->Cells [3][0] = "String";

    // Populate structure grid
    int strucCount = m_vm.DataTypes ().Structures ().size ();
    SetRowCount (StrucGrid, strucCount);
    int i;
    for (i = 0; i < strucCount; i++) {
        vmStructure& s = m_vm.DataTypes ().Structures () [i];
        StrucGrid->Cells [0] [i + 1] = s.m_name.c_str ();
        StrucGrid->Cells [1] [i + 1] = IntToStr (s.m_dataSize);
        StrucGrid->Cells [2] [i + 1] = s.m_containsString ? "Y" : "";
        StrucGrid->Cells [3] [i + 1] = s.m_containsArray  ? "Y" : "";
    }

    // Populate variable grid
    int varCount = m_vm.Variables ().Variables().size ();
    SetRowCount (VarGrid, varCount);
    for (i = 0; i < varCount; i++) {
        vmVariable& v = m_vm.Variables ().Variables () [i];
        VarGrid->Cells [0] [i + 1] = m_vm.DataTypes ().DescribeVariable (v.m_name, v.m_type).c_str ();
        VarGrid->Cells [1] [i + 1] = m_vm.VarToString (v, VM_DATATOSTRINGMAXCHARS).c_str ();
    }

    // Populate code grid
    int count = m_vm.InstructionCount();
    SetRowCount (CodeGrid, count);
    for (i = 0; i < count; i++) {
        vmInstruction ins = m_vm.Instruction (i);
        CodeGrid->Cells [0][i + 1] = IntToStr (i);
        CodeGrid->Cells [1][i + 1] = vmOpCodeName (ins.m_opCode).c_str ();
        CodeGrid->Cells [2][i + 1] = vmBasicValTypeName (ins.m_type).c_str ();
        switch (ins.m_opCode) {
        case OP_DECLARE:
        case OP_LOAD_VAR: {
            int index = ins.m_value.IntVal ();
            if (m_vm.Variables ().IndexValid (index))
                CodeGrid->Cells [3] [i + 1] = m_vm.Variables ().Variables () [index].m_name.c_str ();
            else
                CodeGrid->Cells [3] [i + 1] = "???";
            break;
        }
        case OP_CALL_FUNC: {
            CodeGrid->Cells [3][i + 1] = (m_comp.FunctionName (ins.m_value.IntVal ()) + " ()").c_str ();
            break;
            }
        case OP_COPY:
        case OP_ALLOC: {
            CodeGrid->Cells [3][i + 1] = m_vm.DataTypes ().DescribeVariable ("", m_vm.GetStoredType (ins.m_value.IntVal ())).c_str ();
            break;
        }
        default:
            CodeGrid->Cells [3][i + 1] = m_vm.BasicValToString (ins.m_value, ins.m_type, true).c_str ();
        }
    }

    // Ensure IP is visible
    if (CodeGrid->TopRow > m_vm.IP () + 1)
        CodeGrid->TopRow = m_vm.IP () + 1;
    else if (CodeGrid->TopRow + CodeGrid->VisibleRowCount < m_vm.IP () + 2)
        CodeGrid->TopRow = m_vm.IP () - CodeGrid->VisibleRowCount + 2;

    // Populate mem dump grid
    int memCount = m_vm.Data ().Size ();
    SetRowCount (MemGrid, memCount);
    for (i = 0; i < memCount; i++) {
        vmValue& val = m_vm.Data ().Data () [i];
        MemGrid->Cells [0] [i + 1] = IntToStr (i);
        MemGrid->Cells [1] [i + 1] = m_vm.ValToString (val, VTP_INT,    false).c_str ();
        MemGrid->Cells [2] [i + 1] = m_vm.ValToString (val, VTP_REAL,   false).c_str ();
        MemGrid->Cells [3] [i + 1] = m_vm.ValToString (val, VTP_STRING, false).c_str ();
    }

    // Populate stack grid
    int stackCount = m_vm.Stack ().Size ();
    SetRowCount (StackGrid, stackCount);
    for (i = 0; i < stackCount; i++) {
        vmValue& val = m_vm.Stack () [i];
        StackGrid->Cells [0] [i + 1] = IntToStr (i);
        StackGrid->Cells [1] [i + 1] = m_vm.ValToString (val, VTP_INT,    false).c_str ();
        StackGrid->Cells [2] [i + 1] = m_vm.ValToString (val, VTP_REAL,   false).c_str ();
        StackGrid->Cells [3] [i + 1] = m_vm.ValToString (val, VTP_STRING, false).c_str ();
    }

    // Registers
    Reg1I->Text = m_vm.ValToString (m_vm.Reg (), VTP_INT,  false).c_str ();
    Reg1R->Text = m_vm.ValToString (m_vm.Reg (), VTP_REAL, false).c_str ();
    Reg1S->Text = m_vm.RegString ().c_str ();
    Reg2I->Text = m_vm.ValToString (m_vm.Reg2 (), VTP_INT,  false).c_str ();
    Reg2R->Text = m_vm.ValToString (m_vm.Reg2 (), VTP_REAL, false).c_str ();
    Reg2S->Text = m_vm.Reg2String ().c_str ();

    UpdateFields (0);
    VariableMemo->Text = VarGrid->Cells [1] [1];
    m_updating = false;
}
//---------------------------------------------------------------------------

void __fastcall TVirtualMachineViewForm::StrucGridSelectCell(
      TObject *Sender, int ACol, int ARow, bool &CanSelect)
{
    if (!m_updating)
        UpdateFields (ARow - 1);
}
//---------------------------------------------------------------------------

void TVirtualMachineViewForm::UpdateFields (int row) {
    if (StrucGrid->Tag <= 0)
        SetRowCount (FieldGrid, 0);
    else {
        vmStructure& s = m_vm.DataTypes ().Structures () [row];
        SetRowCount (FieldGrid, s.m_fieldCount);
        for (int i = 0; i < s.m_fieldCount; i++) {
            vmStructureField f = m_vm.DataTypes ().Fields () [s.m_firstField + i];
            FieldGrid->Cells [0] [i + 1] = m_vm.DataTypes ().DescribeVariable (f.m_name, f.m_type).c_str ();
            int dataSize;
            if (i < s.m_fieldCount - 1) dataSize = m_vm.DataTypes ().Fields () [s.m_firstField + i + 1].m_dataOffset - f.m_dataOffset;
            else                        dataSize = s.m_dataSize - f.m_dataOffset;
            FieldGrid->Cells [1] [i + 1] = IntToStr (dataSize);
        }
    }
}
//---------------------------------------------------------------------------

void __fastcall TVirtualMachineViewForm::VarGridSelectCell(TObject *Sender,
      int ACol, int ARow, bool &CanSelect)
{
    if (!m_updating)
        VariableMemo->Text = VarGrid->Cells [1] [ARow];
}
//---------------------------------------------------------------------------

void __fastcall TVirtualMachineViewForm::CodeGridDrawCell(TObject *Sender,
      int ACol, int ARow, TRect &Rect, TGridDrawState State)
{
    TCanvas *c = CodeGrid->Canvas;
    if (ARow == 0) {
        c->Brush->Color = clBtnFace;
        c->Font->Color = clBtnText;
    }
    else if (ARow - 1 == m_vm.IP ()) {
        c->Brush->Color = clRed;
        c->Font->Color = clWhite;
    }
    else {
        c->Brush->Color = clWindow;
        c->Font->Color = clWindowText;
    }
    c->FillRect(Rect);
    c->TextRect (Rect, Rect.left + 2, Rect.top + 2, CodeGrid->Cells [ACol] [ARow]);
}
//---------------------------------------------------------------------------

