/*  Main.cpp

    Copyright (C) 2004, Tom Mulgrew (tmulgrew@slingshot.co.nz)

    Simplified, cut down Basic4GL compiler
        * Compiles to console application
        * No UI
        * No OpenGL or other libraries
        * Only two functions (print & printr)

	Should be quite portable.
*/

#include "TomComp.h"
#include <iostream.h>

// Test program
char buffer [1024];

// Simple "print" function

void WrapPrint (TomVM& vm) {

	// Regular print
	cout << vm.GetStringParam (1).c_str ();
}

void WrapPrintr (TomVM& vm) {

	// Print with newline
	cout << vm.GetStringParam (1).c_str () << endl;
}

int main (void) {

	cout << "TomBasicCore BASIC compiler and virtual machine (C) Tom Mulgrew 2003-2004" << endl;
	cout << "Enter Basic program then hit Ctrl+Z (On Windows)"<< endl;
	cout << "Or pipe a text file into stdin" << endl;

	// Create compiler and virtual machine
	TomVM				vm;
	TomBasicCompiler	comp (vm);

	// Register function wrappers
    comp.AddFunction ("print",  WrapPrint,  compParamTypeList () << VTP_STRING, false, false,  VTP_INT);
    comp.AddFunction ("printr", WrapPrintr, compParamTypeList () << VTP_STRING, false, false,  VTP_INT);

	// Read source from cin
	cin.getline (buffer, 1023);
	while (!cin.fail ()) {
		comp.Parser ().SourceCode ().push_back (buffer);
		cin.getline (buffer, 1023);
	};

	// Compile program
	comp.ClearError ();
	comp.Compile ();
	if (comp.Error ()) {
		cout << endl << "COMPILE ERROR!: " << comp.GetError ().c_str () << endl;
		return 0;
	}

	// Run program
	vm.Reset ();
	do {

		// Execute 1000 VM op-codes (maximum)
		vm.Continue (1000);

	} while (!(vm.Error () || vm.Done ()));

	// Check for virtual machine error
	if (vm.Error ())
		cout << endl << "RUNTIME ERROR!: " << vm.GetError ().c_str () << endl;
	else
		cout << endl << "Done!" << endl;

	return 0;
};
