// kernel32.lib user32.lib gdi32.lib winspool.lib comdlg32.lib advapi32.lib shell32.lib ole32.lib oleaut32.lib uuid.lib odbc32.lib 

/*	Basic4GL standalone stub.
	Copyright (C) Tom Mulgrew, 2003-2004 (tmulgrew@slingshot.co.nz)

	Contains:
		* Virtual machine
		* Compiler
		* Basic4GL program

	Eventually will replace the Basic4GL source code with virtual machine raw data, and
	remove the compiler component completely.

	The Basic4GL data is placed in a special segment named "Basic4GL" at the end of the
	executable.
	This simplifies the exe creation code, as it only needs to update the one section
	(and adjust some file size variables).
*/

#include "TomComp.h"
#include "glTextGrid.h"
#include "TomOpenGLBasicLib.h"
#include "glBasicLib_gl.h"
#include "glbasiclib_glu.h" 
#include "TomStdBasicLib.h"
#include "TomWindowsBasicLib.h"
#include "TomTextBasicLib.h"
#include "TomJoystickBasicLib.h"
#include "TomTrigBasicLib.h"
#include "TomFileIOBasicLib.h"
//#include "TomNetBasicLib.h"
#include "glSpriteEngine.h"
#include "EmbeddedFiles.h"
#include "Standalone.h"

// Allocate space for Basic4GL program
// We only allocate 1 byte, as patcher will resize the program anyway.
#pragma data_seg("Basic4GL")
#ifdef _DEBUG
	char __declspec(allocate("Basic4GL")) basic4GLData[100000];
#else
	char __declspec(allocate("Basic4GL")) basic4GLData[sizeof (StandaloneHeader)];
#endif
#pragma data_seg()

#ifdef _DEBUG
void CreateTestData ();
#endif

int APIENTRY WinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPSTR     lpCmdLine,
                     int       nCmdShow)
{

#ifdef _DEBUG
	CreateTestData ();
#endif
	
	// Allocate objects
    TomVM               vm;											// Virtual machine		
    TomBasicCompiler    comp (vm);									// Compiler
	FileOpener			files;

	// Unpack standalone data
	StandaloneHeader header;
	if (!UnpackStandaloneData (basic4GLData, header, comp.Parser (), files)) {
		MessageBox (NULL, "Program integrity check failed", "Program integrity error", MB_OK);
		return 0;
	}

	// Create OpenGL window
	int xSize, ySize, bpp;
	switch (header.screenMode) {
	case 1:		xSize = 640;	ySize = 480;	break;
	case 2:		xSize = 800;	ySize = 600;	break;
	case 3:		xSize = 1024;	ySize = 768;	break;
	case 4:		xSize = 1280;	ySize = 1024;	break;
	default:	
		xSize = GetSystemMetrics (SM_CXSCREEN);
		ySize = GetSystemMetrics (SM_CYSCREEN);
	}
	switch (header.colourDepth) {
	case 1:		bpp	= 16; break;
	case 2:		bpp = 32; break;
	default:	bpp = 0;
	}
    glTextGridWindow	glWin (header.fullScreen, header.border, xSize, ySize, bpp, header.title, RGM_RECREATEWINDOW);		// Create window (with text support)

	// Initialise them
    glSpriteEngine		glText("charset.png", &files, 25, 40, 16, 16);	// Setup a text grid for the window
    glWin.SetTextGrid (&glText);									
	if (glText.Error ()) {
		MessageBox (glWin.GetHWND (), glText.GetError ().c_str (), "ERROR!", MB_OK);
		return 0;
	}

	// Initialise function and constant libraries
    InitTomStdBasicLib (comp);							// Standard library
    InitTomWindowsBasicLib (comp, &files);				// Windows specific library
    InitTomOpenGLBasicLib (comp, &glWin, &files);		// OpenGL
    InitTomTextBasicLib (comp, &glWin, &glText);
    InitGLBasicLib_gl (comp);
    InitGLBasicLib_glu (comp);
	InitTomJoystickBasicLib (comp, &glWin);
    InitTomTrigBasicLib (comp);							// Trigonometry library
    InitTomFileIOBasicLib (comp, &files);				// File I/O library
//	InitTomNetBasicLib (comp);

	// Compile program
    comp.ClearError ();
    comp.Compile ();
    if (comp.Error ()) {
		MessageBox (glWin.GetHWND (), ((std::string) "Compile error: " + comp.GetError () + ", line " + IntToString (comp.Line () + 1) + ", col " + IntToString (comp.Col () + 1)).c_str (), "ERROR!", MB_OK);
		return 0;
	}

	// Display and initialise OpenGL window
	glWin.Activate ();
	glWin.OpenGLDefaults ();

	// Reset virtual machine
    vm.Reset ();

	try {
		do {
	
			// Execute maximum of 1000 virtual machine Op-codes
			vm.Continue (1000);
	
			// Process windows messages (to keep application responsive)
			glWin.ProcessWindowsMessages();
	
	    } while (!vm.Error () && !vm.Done () && (!(glWin.Closing () && header.escKeyQuits)));
	}
	catch (...) {
		vm.MiscError ("An exception occured");
	}
	
	if (vm.Error ()) {

		// Runtime error encountered.
		// Perform whatever error logic was requested in header file.
		if (header.fullScreen) 
			glWin.Hide ();

		if (header.errorOption == 1)
	
			// Show generic error message
			MessageBox (glWin.GetHWND (), "An error has occured.", "ERROR!", MB_OK);
		else if (header.errorOption == 2) {

			// Show specific error message
			int line, col;
			vm.GetIPInSourceCode (line, col);
			MessageBox (glWin.GetHWND (), ((std::string) "Runtime error: " + vm.GetError () + ", line " + IntToString (line + 1) + ", col " + IntToString (col + 1)).c_str (), "ERROR!", MB_OK);
		}
		// Otherwise just exit
	}
	else if (vm.Done ()) {
		
		// Program completed normally
		// Perform whatever close down logic was requested in header file.
		// (Wait for key, just close e.t.c)

		// Invalid case, fullscreen + wait for window to be closed.		
		if (header.fullScreen && header.finishedOption == 2)
			header.finishedOption = 1;			// Replace with wait for keypress

		if (header.finishedOption == 1) {
			// Wait for keypress or close window
			glWin.ProcessWindowsMessages();
			while (glWin.GetKey () != 0 && !glWin.Closing ()) glWin.ProcessWindowsMessages();
			while (glWin.GetKey () == 0 && !glWin.Closing ()) glWin.ProcessWindowsMessages();
		}
		else if (header.finishedOption == 2) {
			// Wait for window to be closed
			while (!glWin.Closing ())
				glWin.ProcessWindowsMessages ();
		}
		// Otherwise just exit
	}	

//	CleanupTomNetBasicLib ();
	return 0;
}

#ifdef _DEBUG
void CreateTestData () {

	// Create data in stream
	std::stringstream s;
	
	// Write standalone header
	StandaloneHeader header;
	header.version			= STANDALONE_VERSION;
	header.fullScreen		= false;
	header.screenMode		= 2;
	header.colourDepth		= 0;
	header.finishedOption	= 1;
	header.errorOption		= 2;
	header.escKeyQuits		= true;
	memcpy (header.title, "Test program\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0", 32);
	s.write ((char *) &header, sizeof (header));

	// Write program data
	std::string program = "\
    dim rtri#\r\n\
    dim rquad#                                        \r\n\
    while true\r\n\
		glDisable (GL_TEXTURE_2D)\r\n\
		glClearColor (1, 0, 0, 1)\r\n\
        glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT)    \r\n\
        glLoadIdentity()                            \r\n\
        glTranslatef(-1.5,0.0,-6.0)                    \r\n\
        glRotatef(rtri#,0.0,1.0,0.0)                    \r\n\
        glBegin(GL_TRIANGLES)                            \r\n\
            glColor3f(1.0,0.0,0.0)                    \r\n\
            glVertex3f( 0.0, 1.0, 0.0)                    \r\n\
            glColor3f(0.0,1.0,0.0)                    \r\n\
            glVertex3f(-1.0,-1.0, 0.0)                    \r\n\
            glColor3f(0.0,0.0,1.0)                    \r\n\
            glVertex3f( 1.0,-1.0, 0.0)                    \r\n\
        glEnd()                                    \r\n\
        glLoadIdentity()                            \r\n\
        glTranslatef(1.5,0.0,-6.0)                        \r\n\
        glRotatef(rquad#,1.0,0.0,0.0)                    \r\n\
        glColor3f(0.5,0.5,1.0)                        \r\n\
        glBegin(GL_QUADS)                            \r\n\
            glVertex3f(-1.0, 1.0, 0.0)                    \r\n\
            glVertex3f( 1.0, 1.0, 0.0)                    \r\n\
            glVertex3f( 1.0,-1.0, 0.0)                    \r\n\
            glVertex3f(-1.0,-1.0, 0.0)                    \r\n\
        glEnd()                                    \r\n\
        rtri# = rtri# + 0.2                            \r\n\
        rquad# = rquad# + 0.15                        \r\n\
        SwapBuffers ()\r\n\
    wend\r\n";  
	
	int size = program.length () + 1;
	s.write ((char *) &size, sizeof (size));
	s.write (program.c_str (), size);

	// Write file
	int fileCount = 2;
	s.write ((char *) &fileCount, sizeof (fileCount));
	EmbedFile ("files\\test.txt", s);
	EmbedFile ("charset.png", s);

	// Copy stream to Basic4GLData area
	size = s.tellp ();
	s.seekg (0, std::ios::beg);
	s.read (basic4GLData, size);
}
#endif