/*	Created 29-Feb-2004: Thomas Mulgrew (tmulgrew@slingshot.co.nz)
	Copyright (C) Tom Mulgrew, 2004.

	Structures and routines for standalone Basic4GL programs.

	Standalone programs have all their necessary data patched into the end of the program
	by the standalone .exe generator.
	This module defines the structure of this data, along with the necessary routines to
	unpack it into a useable format.

*/

#include "standalone.h"

bool UnpackStandaloneData (char *rawData, StandaloneHeader& header, compParser& parser, FileOpener& files) {
	int offset = 0;

	// Read header
	memcpy (&header, rawData, sizeof (header));
	offset += sizeof (header);

	// Check header validity
	if (header.version != STANDALONE_VERSION)
		return false;

	// Read program code
	int programSize = *((int *) (rawData + offset));
	offset += sizeof (programSize);	
	std::string program = (rawData + offset);
	offset += programSize;

	// Load program into parser, line by line
    parser.SourceCode ().clear ();				// Clear the source code
	int sourceOffset = 0, 
		length = program.length ();
	while (sourceOffset < length) {

		// Look for next newline
		int eol = program.find ('\r', sourceOffset);
		if (eol == std::string::npos) 
			eol = length;

		// Extract line
		std::string line = program.substr (sourceOffset, eol - sourceOffset);
		sourceOffset = eol + 1;

		// Feed into parser
		parser.SourceCode().push_back (line);
	}

	// Remaining data is embedded files
	files.AddFiles (rawData, offset);
	return true;
}
